/**
 ** pif_DiceSimulator
 **   Version: 1.1.20191223
 **   Release Date: December 23, 2019
 **
 ***************************************************************************************************
 ***************************************************************************************************

pif_DiceSimulator is a library that can be used to simulate the rolls of dice with both
arbitrarily-many sides and arbitrarily-numbered sides, as well as the rolls of collections
of these dice.

 +-------------------------------------------+
 |                                           |
 |  Contents                                 |
 |  1. License                               |
 |  2. Introduction                          |
 |  3. Comparison to Built-In Functionality  |
 |  4. How to Use                            |
 |     4.1. Basic Dice                       |
 |          4.1.1. Common Dice               |
 |     4.2. Custom Dice                      |
 |     4.3. Collections of Dice              |
 |  5. Preprocessors                         |
 |  6. /pif_DiceSimulator/dice Class         |
 |  7. /pif_DiceSimulator/Collection Class   |
 |  8. Release Notes                         |
 |                                           |
 +-------------------------------------------+

  1. License
  ----------------------------------------------------

  The code for pif_DiceSimulator is released under the MIT License. Refer to "+License.dm" for the
  full text of the license.

  2. Introduction
  ----------------------------------------------------

  pif_DiceSimulator is a library that can be used to go beyond the native functionality of the DM
  language in terms of simulation of dice rolls. Its intended use is to allow programmers to
  implement non-standard dice of various sorts, and to let them simulate rolling them either
  individually or as collections of dice. Furthermore, as collection of die one can maintain a state
  of rolled dice for arbitrarily long and perform inspection and filtration of these die.

  3. Comparison to Built-In Functionality
  ----------------------------------------------------

  The primary built-in functions that one would use to simulate dice rolls are roll(), pick(), and
  rand().

  roll() is limited, in that while it is able to simulate the rolling of arbitrarily many die with
  arbitrarily-many sides, these die are always 'standard'. That is, the die it simulates are always
  numbered with {1, ..., k} where k >= 1 is some natural number. Thus, it can not simulate biased
  die or coins, non-transitive die, or any other interesting die that result from uneven
  probabilities. Furthermore, the result of multiple die is always summed and output, precluding any
  sort of filtration or inspection of individual die.

  pick() can mitigate these issues by allowing for the simulation of non-standard die, but it does
  not allow for the easy inspection and filtration of die without what would  be, effectively, re-
  implementing this library. rand() is similar in this regard, in that while this library uses
  rand() for its random numbers, to mimic this library's functionality you would effectively be re-
  implementing the library.

  4. How to Use
  -----------------------------------------------------

  This library comes with two parts: the simulation of individual dice, and the handling of
  collections of die.

     4.1. Basic Dice
     --------------------------------------------------

     Basic dice refers to any dice that can be stored as a list of sides. For example, standard die
     such as a six-sided die could be stored as list(1, 2, 3, 4, 5, 6), while a die with a slight
     bias for 1 and a slightly stronger bias for 6 could be stored as list(1, 1, 2, 3, 4, 5, 6, 6,
     6).

     These sorts of die can be implemented without any extension of the native classes. To do this,
     one must simply pass the list as either a /list object or a list of arguments to the
     constructor of the /pif_DiceSimulator/dice class. For example, a standard six-sided die could
     be implemented on the fly as

         var/pif_DiceSimulator/dice/d6 = new(1, 2, 3, 4, 5, 6)

     Then, to simulate a dice roll, one would simply use the /dice.Roll() method:

         world << d6.Roll() // Would output 1, 2, 3, 4, 5, or 6.

       4.1.1. Common Dice
       ------------------------------------------------

       To simplify things, seven common types of dice (including all dice derived from Platonic
       solids) are pre-implemented as children of the /dice class.

         1. A two-sided die (a.k.a. a coin), under /pif_DiceSimulator/d2 and with an alias under
            /pif_DiceSimulator/coin (technically a child of /d2 without any changes). In addition
            to the standard /dice.Roll() method, it also has a /d2.Flip() method which is an alias
            for Roll(). Additionally, the Map() method will output "heads" and "tails".
         2. A four-sided die, implemented under /pif_DiceSimulator/d4.
         3. A six-sided die (the prototypical type of die), implemented under /pif_DiceSimulator/d6.
         4. An eight-sided die, under /pif_DiceSimulator/d8.
         5. A ten-sided die, under /pif_DiceSimulator/d10.
         6. A twenty-sided die, under /pif_DiceSimulator/d20.
         7. A fudge die, used in the Fudge role-playing system. This die has numeric sides given by
            [-1, -1, 0, 0, 1, 1] when using the Roll() method, and text sides given by ["-", "-",
            " ", " ", "+", "+"] when using the Map() method. See Section 3.2. and/or Section 5. for
            more detail on these methods.

     4.2. Custom Dice
     --------------------------------------------------

     In addition to simple dice, one may also extend the /pif_DiceSimulator/dice class in order to
     create die with more complex behaviors. To do this, one would create a child class of the /dice
     class and modify its Roll() method and possibly its Map() method.

     The Roll() method should *only* produce numeric outputs, or else functionality of this library
     will break. If you wish to associate a text (or other) value with the sides, the Map() method
     should be used instead. When this method is not supplied an argument, it will simulate a dice
     roll (by calling the Roll() method) and then outputs the associated value of the simulated
     roll.

     4.3. Collections of Dice
     --------------------------------------------------

     While simulating dice rolls is the first part of this library, the second part is the handling
     of rolling collections of various dice. This is done by the /pif_DiceSimulator/Collection
     class. This class is instantiated in one of the following two ways.

         new /pif_DiceSimulator/Collection(D1, ... ,Dk)
         new /pif_DiceSimulator/Collection("NdS")

     In the first, D1 through Dk are /dice objects that will be simulated being rolled. In the
     latter, "NdS" is a string where N is an optional, positive natural number indicating the number
     of dice to roll, and S is a positive natural number indicating the number of sides on the die
     being rolled. If N is left off (i.e., the string is "dS" instead of "NdS") it defaults to 1.

     When you wish to roll a collection of dice, you call the Roll() method. This method will, as it
     simulates the roll, order the data from smallest to largest based on the numerical value of the
     roll, and also generate a list of data pulled from Map() for each roll, arranged using the
     order data of the numeric values. This data is stored until it is cleared either by another
     call to Roll() or by calling the Clear() method.

     The /Collection class also has several filtration methods, allowing you to sort out rolls that
     don't meet certain criteria. These are Above(), Below(), Inside(), and Outside():

       1. Above() returns only rolls that are at or above a certain bound.
       2. Below() returns only rolls that are at or below a certain bound.
       3. Inside() returns only rolls that are within an upper and lower bound, inclusive.
       4. Outside() returns only rolls that are above a certain upper bound or below a certain lower
          bound, inclusive.

     There is also MapAbove(), MapBelow(), MapInside(), and MapOutside() for the Map() values of
     these rolls.

     Both the inspection and filtration methods are useful because they result in much different
     probability distributions that you would get from just taking the sum of the die, though this
     is still possible via the Sum() method. Thus, they allow you to skew results in favor of one
     outcome or another by choosing die in a certain manner--though not with absolute certainty. In
     terms of game design, this can make for interesting choices for the player.

  5. Preprocessors
  -----------------------------------------------------

  By default, the namespace of this library is prefixed, i.e. /pif_DiceSimulator has the prefix _pif
  on it. This can be removed by either,

    1. Defining PIF_NOPREFIX_DICESIMULATOR, which will remove the prefix from only this library.
    2. Defining PIF_NOPREFIX_GENERAL, which will remove the prefix from this library and all others
       of mine that in the future will have similar prefixing.

  As expected, the unprefixed namespace is /DiceSimulator.

  There is additionally a PIF_DICESIMULATOR_INCLUDED flag defined, to allow for the optional
  inclusion of code depending on the presence of this library.

  6. /pif_DiceSimulator/dice Class
  -----------------------------------------------------

    +---------------------------------+
    |  Parent Class: /datum           |
    |  Namespace: /pif_DiceSimulator  |
    |  File: Dice class.dm            |
    +---------------------------------+

  The /dice class is the main class used in simulating particular die. Its methods are as follows:

    New()
    --------------------

    Arguments.

      Arguments for the constructor are allowed in one of two formats.

        1. new /pif_DiceSimulator/dice(d_1, ..., d_k)
        2. new /pif_DiceSimulator/dice( list(d_1, ..., d_k) )

      In both cases, d_1 through d_k are numeric value.

    Behavior.

      Given the arguments d_1, ..., d_k, the constructor will create a new k-sided die with sides
      populated from d_1, ..., d_k. E.g., if the arguments were (1, 2, 3, 4, 5, 6), the die being
      simulated would be a six-sided die with sides 1, 2, 3, 4, 5, and 6. That is, it would be a
      standard six-sided die.

      If any of the arguments are non-numeric, the currently-generated data is deleted and an
      exception is thrown.

    Returns.

      The method returns 1 on success.

    Map()
    --------------------

    Arguments.

      An optional numeric argument. This argument, if present, should be one of the sides of the
      dice.

    Behavior.

      Two possible behaviors.

        1. If an argument was passed, the method outputs an "associated value" to that argument,
           which should represent what the given side "means". For example, if a coin-simulating die
           has sides 0 and 1, 0 might map to "tails" while 1 might map to "heads".

           If the argument is not recognized as one of the legal values, null is returned.

        2. If no argument was passed, the method calls Roll() and then returns the value associated
           to the value of Roll().

    Returns.

      Non-null value on success. Null on failure, where failure indicates the passed argument was
      not recognized.

    Note: Map() in the /dice class returns the value of Roll() when passed no argument, and the
          argument itself when passed any other value. This behavior is to help the implementation
          of Roll() in subclasses, but it effectively does nothing in /dice itself.

    Maximum()
    --------------------

    Returns.

      Numeric value of the largest possible.

    Minimum()
    --------------------

    Returns.

      Numeric value of the smallest possible side.

    Roll()
    --------------------

    Behavior.

      Uniformly outputs from the available sides. Note that this just means the choice from the
      possible sides is done uniformly, *not* that the sides themselves are uniform. For example, if
      your die has sides {1, 1, 2}, then while it will choose which of the three values to output in
      a uniform manner, 1 will nevertheless occur more often because it is an option more often.

    Returns.

      Numeric value of the side that was rolled.

    Sides()
    --------------------

    Returns.

      A copy of the list of sides.

    Size()
    --------------------

    Returns.

      The number of sides the die has.

  7. /pif_DiceSimulator/Collection Class
  -----------------------------------------------------

    +---------------------------------+
    |  Parent Class: /datum           |
  	|  Namespace: /pif_DiceSimulator  |
    |  File: Collection class.dm      |
    +---------------------------------+

  The /Collection class is used to handle the rolling of collections of dice.

    New()
    --------------------

    Arguments.

      There are four legal formats for arguments.

        1. new /pif_DiceSimulator/Collection(D1, ..., Dk)
        2. new /pif_DiceSimulator/Collection( list(D1, ..., Dk) )
        3. new /pif_DiceSimulator/Collection("NdS")
        4. new /pif_Dicesimulator/Collection("dS")

      In 1. and 2., D1 through Dk are /pif_DiceSimulator/dice object or objects of children of that
      class. If any of these arguments do not fit this criteria, and exception is thrown. Note that
      D1 through Dk need not all be unique. E.g., if you wish to have multiple of one type of die,
      you can pass the same object multiple times.

      In 3, and 4. "NdS" and "dS" are strings where N and S are positive natural numbers. If N is
      left off (i.e., format 4. is used), then it defaults to one. If N or S are either non-positive
      or fractional, an exception is thrown.

    Behavior.

      In formats 1. and 2., the constructor simple stores the /dice objects.

      In formats 3. and 4., the constructor generates a new /dice object with sides labelled {1,
      ..., S} and then stores that object N times.

    Returns.

      1 on success.

    Above()
    --------------------

    Arguments.

      A numeric value.

    Returns.

      A /list of rolled values greater than or equal to the argument.

    At()
    --------------------

    Arguments.

      There are two possibilities.

        1. An integer between 1 and K.
        2. An integer between -1 and -K.

      Where K is the total number of dice being rolled.

    Behavior.

      When the argument is positive, the At() method looks up the ordered value of the roll in the
      position of the argument. For example, if the argument is 1, it looks up the roll that had the
      smallest numerical value, while if it was K it looks up the roll with the highest numerical
      value.

      When the argument is negative, it looks it up in decreasing order. For example, -1 indicates
      the highest value while -K indicates the smallest value.

    Returns.

      The numeric value of the roll in the position the argument indicates.

    Below()
    --------------------

    Arguments.

      A numeric value.

    Returns

      A /list of rolled values less than or equal to the argument.

    Clear()
    --------------------

    Behavior.

      Clears the stored data for the previous roll.

    Count()
    --------------------

    Returns

      The total number of dice being rolled.

    Inside()
    --------------------

    Arguments.

      1. A numeric lower bound.
      2. A numeric upper bound.

    Returns

      A /list of rolled values within those bounds (inclusively).

    Largest()
    --------------------

    Returns

      The largest-valued result rolled.

    Note: Alias for At(-1)

    Map()
    --------------------

    Returns

      A /list of the Map() values for the roll.

    MapAbove()
    --------------------

    Arguments

      A numeric upper bound.

    Returns

      A /list of Map() values for rolls whose numeric value is greater than or equal to the
      upper bound.

    MapAt()
    --------------------

    Arguments.

      There are two possibilities.

        1. An integer between 1 and K.
        2. An integer between -1 and -K.

      Where K is the total number of dice being rolled.

    Behavior.

      When the argument is positive, the MapAt() method looks up the Map() value of ordered value of
      the roll in the position of the argument. For example, if the argument is 1, it looks up the
      Map() value of the roll that had the smallest numerical value, while if it was K it looks up
      the Map() value of the roll with the highest numerical value.

      When the argument is negative, it looks it up in decreasing order. For example, -1 indicates
      the Map() of highest value while -K indicates the Map() of the smallest value.

    Returns.

      The Map() of the numeric value of the roll in the position the argument indicates.

    MapBelow()
    --------------------

    Arguments

      A numeric lower bound.

    Returns

      A /list of Map() values for rolls whose numeric value is less than or equal to the lower
      bound.

    MapInside()
    --------------------

    Arguments.

      1. A numeric lower bound.
      2. A numeric upper bound.

    Returns.

      A /list of Map() values for rolls whose numeric value is above the lower bound and below the
      upper bound (inclusive).

    MapLargest()
    --------------------

    Returns

      The Map() value of the largest roll.

    Note: Alias for MapAt(-1)

    MapOutside()
    --------------------

    Arguments.

      1. A numeric lower bound.
      2. A numeric upper bound.

    Returns.

      A /list of Map() values for rolls whose numeric value is below the lower bound and above the
      upper bound (inclusive)

    MapSmallest()
    --------------------

    Returns

      The Map() value of the smallest roll.

    Note: Alias for MapAt(1)

    Outside()
    --------------------

    Arguments.

      1. A numeric lower bound.
      2. A numeric upper bound.

    Returns.

      A /list of rolls whose numeric value is below the lower bound and above the upper bound
      (inclusively).

    Result()
    --------------------

    Returns.

      A /list of values rolled.

    Roll()
    --------------------

    Behavior.

      Rolls each die and stores both their value and the output of their Map() of that value, and
      arranges these results in order according to the roll of that die.

      This method will also call Clear() for the previous rolls.

    Smallest()
    --------------------

    Returns.

      The numeric value of the smallest roll.

    Note: Alias for At(1).

    Sum()
    --------------------

    Returns.

      The numeric value of the sum of the roll results.

  8. Release Notes
  -----------------------------------------------------

  Version 1.1.20191223

    - Changed class naming conventions to be consistent with future libraries (capitalized the first
      letter of class names).

    - Updated documentation.
      - Changed class naming convention to match class implementation.
      - Added some extra class information.
      - Documented the PIF_DICESIMULATOR_INCLUDED preprocessor.
      - Added Release Notes section.

    Note: Apparently this update has been sitting on my hard drive since late October 2015, and I
          never actually got around to putting it up on the hub. Whoops!

  Version 1.0.20151016.

    - Initial release.

 ***************************************************************************************************
 ***************************************************************************************************/