/*
 * The /pif_DiceSimulator/collection class simulates rolling a collection of various dice, as
 * specified by the programmer. As these dice are rolled, they are stored in numerically-increasing
 * order, allowing for the simple selection of elements by size.
 */

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
pif_DiceSimulator
#else
DiceSimulator
#endif
	Collection
		parent_type = /datum

		var
			list
				// The /pif_DiceSimulator/Dice objects passed to the instantiator.
				_dice

				// The current rolls, and their map value.
				_rolls
				_maps

		New(...)
			// This can be instantiated either as
			//   new(d1, ..., dk)
			//   new( list(d1, ..., dk) )
			//
			//   new("KdS")
			//   new("dS")
			// where,
			//   - d1 through dk must be /pif_DiceSimulator/Dice objects.
			//   - K is the number of dice to roll. Must be a positive natural number, and defaults
			//     to 1 if left off.
			//   - S is the number of sides for the dice. Must be a positive natural number.

			_dice = new

			if(args.len > 1)
				// When the die are passed as a list of arguments.
				for(var/x in args)
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
					if(!istype(x, /pif_DiceSimulator/Dice))
#else
					if(!istype(x, /DiceSimulator/Dice))
#endif
						_dice = null
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
						throw EXCEPTION("/pif_DiceSimulator/collection only accepts /pif_DiceSimulator/Dice objects for constructor argument objects.")
#else
						throw EXCEPTION("/DiceSimulator/collection only accepts /DiceSimulator/Dice objects for constructor argument objects.")
#endif

					else
						_dice += x

			else if(args.len == 1)
				// When only one argument has been passed, there are four possibilities.

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
				if(istype(args[1], /pif_DiceSimulator/Dice))
#else
				if(istype(args[1], /DiceSimulator/Dice))
#endif
					// The user wants a collection of one die.
					_dice += args[1]

				else if(istype(args[1], /list))
					// That they passed a list of die.
					for(var/x in args[1])
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
						if(!istype(x, /pif_DiceSimulator/Dice))
#else
						if(!istype(x, /DiceSimulator/Dice))
#endif
							_dice = null
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
							throw EXCEPTION("/pif_DiceSimulator/collection only accepts /pif_DiceSimulator/Dice objects for constructor argument objects.")
#else
							throw EXCEPTION("/DiceSimulator/collection only accepts /DiceSimulator/Dice objects for constructor argument objects.")
#endif

						else
							_dice += x

				else if(istext(args[1]))
					// That they passed the argument as an "NdS" or "dS" string.
					var
						pos = findtextEx(args[1], "d")

						// As with the argument format, N is the number of dice to roll and
						// S is the number of sides on the die.
						N
						S

						list/sides = new

					if(pos == 0)
						_dice = null
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
						throw EXCEPTION("/pif_DiceSimulator/collection constructor string must be of the form \"NdS\" or \"dS\".")
#else
						throw EXCEPTION("/DiceSimulator/collection constructor string must be of the form \"NdS\" or \"dS\".")
#endif

					// Parsing the incoming string and verifying its data is valid.

					S = text2num(copytext(args[1], pos+1))

					if(pos == 1)
						N = 1
					else
						N = text2num(copytext(args[1], 1, pos))

					if((N == null) || (S == null))
						_dice = null
						throw EXCEPTION("Constructor string malformed.")

					if(S == 0)
						throw EXCEPTION("Dice must have positive number of sides.")
					else if(round(S) != S)
						throw EXCEPTION("Dice must have non-fractional number of sides.")

					if(N == 0)
						throw EXCEPTION("Must be positive number of dice in collection.")
					else if(round(N) != N)
						throw EXCEPTION("Must be non-fractional number of dice in collection.")

					// Actually populating the collection.

					for(var/i in 1 to S)
						sides += i

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
					var/pif_DiceSimulator/Dice/die = new(sides)
#else
					var/DiceSimulator/Dice/die = new(sides)
#endif
					for(var/i in 1 to N)
						_dice += die

				else
					// Or that they did something wrong.
					_dice = null
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
					throw EXCEPTION("/pif_DiceSimulator/collection only accepts /pif_DiceSimulator/Dice objects for constructor argument objects.")
#else
					throw EXCEPTION("/DiceSimulator/collection only accepts /DiceSimulator/Dice objects for constructor argument objects.")
#endif

			else
				// Nothing passed at all.
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
				throw EXCEPTION("Must pass /pif_DiceSimulator/Dice objects or string of the form \"NdS\" or \"dS\" as arguments to /pif_DiceSimulator/collection constructor.")
#else
				throw EXCEPTION("Must pass /DiceSimulator/Dice objects or string of the form \"NdS\" or \"dS\" as arguments to /DiceSimulator/collection constructor.")
#endif

			return 1

		proc
			Count()
				// The total number of dice being rolled.
				return _dice.len

			Roll()
				// This rolls all the die.

				Clear()

				_rolls = new
				_maps = new
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
				for(var/pif_DiceSimulator/Dice/dice in _dice)
#else
				for(var/DiceSimulator/Dice/dice in _dice)
#endif
					var/r = dice.Roll()

					var/i
					for(i = 1, i <= _rolls.len, i ++)
						if(_rolls[i] > r)
							_rolls.Insert(i, r)
							_maps.Insert(i, dice.Map(r))
							break

					if(i >= _rolls.len)
						_rolls += r
						_maps += dice.Map(r)

				return Result()

			Result()
				// A duplicate of the _rolls list.
				return _rolls.Copy()

			Map()
				// A duplicate of the _maps list.
				return _maps.Copy()

			Clear()
				// Clears the stored data.
				_rolls = null
				_maps = null

				return 1

			At(i)
				// Returns the roll at a specified position. If i is positive, it counts from the
				// beginning, while if it's negative it counts from the end. The distributions of
				// these individual values can be found via order statistic methods.

				if(!isnum(i))
					throw EXCEPTION("Roll position must be numeric.")
				else if(round(i) != i)
					throw EXCEPTION("Roll position must be integral.")
				else if(i == 0)
					throw EXCEPTION("Roll position must be non-zero.")
				else if(abs(i) > Count())
					throw EXCEPTION("Roll position must be in range of data.")

				if(i > 0)
					return _rolls[i]
				else if(i < 0)
					return _rolls[Count() + 1 + i]

			MapAt(i)
				// Same as At(), but for the Map() value.

				if(!isnum(i))
					throw EXCEPTION("Roll position must be numeric.")
				else if(round(i) != i)
					throw EXCEPTION("Roll position must be integral.")
				else if(i == 0)
					throw EXCEPTION("Roll position must be non-zero.")
				else if(abs(i) > Count())
					throw EXCEPTION("Roll position must be in range of data.")

				if(i > 0)
					return _maps[i]
				else if(i < 0)
					return _maps[Count() + 1 + i]

			Largest()
				// Max of the rolls. An alias for At(-1).
				return At(-1)

			Smallest()
				// Min of the rolls. An alias for At(1).
				return At(1)

			MapLargest()
				// The Map() value of the die at the largest position. Alias of MapAt(-1)
				return MapAt(-1)

			MapSmallest()
				// The Map() value of the die at the smallest position. Alias of MapAt(1)
				return MapAt(1)

			Sum()
				// The sum of the roll.
				. = 0
				for(var/i in _rolls)
					. += i

			Above(bound)
				// The rolls greater than or equal to bound.
				. = list()
				for(var/i in _rolls)
					if(i >= bound)
						. += i

			MapAbove(bound)
				// Map() equivalent of Above()
				. = list()
				for(var/i in 1 to Count())
					if(_rolls[i] >= bound)
						. += _maps[i]

			Below(bound)
				// The rolls less than or equal to bound.
				. = list()
				for(var/i in _rolls)
					if(i <= bound)
						. += i

			MapBelow(bound)
				// Map() equivalent of Below()
				. = list()
				for(var/i in 1 to Count())
					if(_rolls[i] <= bound)
						. += _maps[i]

			Inside(lower, upper)
				// The rolls between lower and upper, inclusive.
				. = list()
				for(var/i in _rolls)
					if((i >= lower) && (i <= upper))
						. += i

			MapInside(lower, upper)
				// Map() equivalent of Inside().
				. = list()
				for(var/i in 1 to Count())
					if((_rolls[i] >= lower) && (_rolls[i] <= upper))
						. += _maps[i]

			Outside(lower, upper)
				// The rolls above upper and below lower, inclusive.
				. = list()
				for(var/i in _rolls)
					if((i <= lower) || (i >= upper))
						. += i

			MapOutside(lower, upper)
				// The Map() equivalent of Outside().
				. = list()
				for(var/i in 1 to Count())
					if((_rolls[i] <= lower) || (_rolls[i] >= upper))
						. += _maps[i]