/*
 * The /pif_DiceSimulator/Dice class simulates a many-sided die (i.e., a die with one or more
 * (naturally-numbered) sides) and with arbitrary numbers on those sides. In particular, the generic
 * /pif_DiceSimulator/Dice class does this generically, but can be subclassed to allow for more
 * complex behavior.
 */

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
pif_DiceSimulator
#else
DiceSimulator
#endif
	Dice
		parent_type = /datum

		var
			// Should not be modified by the programmer..
			list/_sides

			_minimum = null
			_maximum = null

			_size

		New(...)
			// This can either be formatted as either,
			//   New(s1, ..., sk)
			//   New( list(s1, ..., sk) )

			_sides = new

			if(args.len > 1)
				// If passed as a list of arguments.
				for(var/x in args)
					if(isnum(x))
						_sides += x

						_minimum = min(x, _minimum)
						_maximum = max(x, _maximum)
					else
						_sides = null
						_minimum = null
						_maximum = null

						throw EXCEPTION("Non-numeric dice side passed.")
			else if((args.len == 1) && !istype(args[1], /list))
				// A 'one-sided die', if they for some reason want it. The geometry
				// of such a shape is intriguing.
				if(isnum(args[1]))
					_minimum = args[1]
					_maximum = args[1]

					_sides += args[1]
				else
					_sides = null
					_minimum = null
					_maximum = null

					throw EXCEPTION("Non-numeric dice side passed.")

			else if(istype(args[1], /list))
				// If passed as a list.
				for(var/x in args[1])
					if(isnum(x))
						_sides += x

						_minimum = min(x, _minimum)
						_maximum = max(x, _maximum)
					else
						_sides = null
						_minimum = null
						_maximum = null

						throw EXCEPTION("Non-numeric dice side passed.")

			_size = _sides.len

			return 1

		proc
			Minimum()
				// The lowest numerical value of the die.
				return _minimum
			Maximum()
				// The highest numerical value of the die.
				return _maximum

			Size()
				// The number of sides of the die.
				return _size
			Sides()
				// Produces a list of the sides of the die.
				return _sides.Copy()

			Roll()
				// Rolls the die once.
				return _sides[rand(1, Size())]

			Map(_result = null)
				// In this class, Map() is effectively just an alias for Roll(). Its intended
				// use is for extending the /Dice class, via one of two behaviors depending
				// on whether an argument was supplied.

				// If an argument is supplied, then this method should output a value 'associated'
				// with that output. For example, if you wanted a die with sides {A, B, C, D},
				// then you could have 1 => "A", 2 => "B", etc.

				// If an argument isn't supplied, this method calls roll, then outputs the associated
				// value of the result.

				return (_result == null) ? Roll() : _result

/*
 * Implements a variety of common types of die. All are inherited from the /Dice
 * class.
 */

 	///
 	/// The following all have uniform distributions. That is, every outcome is equally-
 	/// likely.
 	///

 	//
 	// A two-sided die, a.k.a. a coin.
 	//

	d2
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 1
		Minimum()
			return 0

		Size()
			return 2
		Sides()
			return list(0, 1)

		Roll()
			return rand(0,1)

		Map(_result = null)
			_result = ..()

			if(_result == 1)
				return "heads"
			else if(_result == 0)
				return "tails"
			else
				return null

		proc/Flip()
			// Alias for Roll(), as it is more natural to want to flip a coin than roll one.
			return Roll()

	coin
		// Alias for coin.
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/d2
#else
		parent_type = /DiceSimulator/d2
#endif

	//
	// A four-sided die.
	//

	d4
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 4
		Minimum()
			return 1

		Size()
			return 4
		Sides()
			return list(1, 2, 3, 4)

		Roll()
			return rand(1, 4)

	//
	// A six-sided die.
	//

	d6
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 6
		Minimum()
			return 1

		Size()
			return 6
		Sides()
			return list(1, 2, 3, 4, 5, 6)

		Roll()
			return rand(1, 6)

	//
	// An eight-sided die.
	//

	d8
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 8
		Minimum()
			return 1

		Size()
			return 8
		Sides()
			return list(1, 2, 3, 4, 5, 6, 7, 8)

		Roll()
			return rand(1, 8)

	//
	// A ten-sided die.
	//

	d10
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 10
		Minimum()
			return 1

		Size()
			return 10
		Sides()
			return list(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

		Roll()
			return rand(1, 10)

	//
	// A twenty-sided die.
	//

	d20
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 20
		Minimum()
			return 1

		Size()
			return 20
		Sides()
			return list( 1,  2,  3,  4,  5,  6,  7,  8,  9, 10,
			            11, 12, 13, 14, 15, 16, 17, 18, 19, 20)

		Roll()
			return rand(1, 20)

	//
	// A Fudge die.
	//

	fudge
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_DICESIMULATOR)
		parent_type = /pif_DiceSimulator/Dice
#else
		parent_type = /DiceSimulator/Dice
#endif

		New()
			return 1

		Maximum()
			return 1
		Minimum()
			return -1

		Size()
			return 6
		Sides()
			return list(-1, -1, 0, 0, 1, 1)

		Roll()
			// While Fudge die are technically six-sided, two sides each are
			// mapped to one outcome, and thus each side has a 1/3 probability.
			// Therefore, this matches that distribution even though it doesn't
			// simulate it exactly.
			return rand(-1, 1)

		Map(_result = null)
			_result = ..()

			if(_result == -1)
				return "-"
			else if(_result == 0)
				return " "
			else if(_result == 1)
				return "+"

			else
				return null